/**
 * @file ContentsStorageBase.cpp
 * ContentsStorageBase class implementation.
 *
 * @author Andrey Lapshov <developers.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#include <unistd.h>

#include "ContentsStorageBase.hpp"

#define DESTRUCTOR_TIME_WAIT 32000

/** ContentsStorageBase::refineExctract method.
 *  Method extracts and returns, from cross-thread storage "refinesVector", first available Refine component
 **/
ContentsStorageBase::RefinePtrType ContentsStorageBase::refineExctract()
{
  RefinePtrType ret = nullptr;
  while(ret.isNull())
  {
    refinesVectorMutex.lock();
    for(unsigned int i = 0; i < refinesVector.size(); i++)
    {
      if(!(refinesVector[i].second))
      {
        ret = refinesVector[i].first;
        refinesVector[i].second = true;
        break;
      }
    }
    refinesVectorMutex.unlock();
    if(ret.isNull())
    {
      componentMutex.lock();
      conditionVar.wait(componentMutex);
      componentMutex.unlock();
    }
  }
  return ret;
}

/** ContentsStorageBase::refineRelease method.
 *  Method releases previously captured Refine component item, in cross-thread storage "refinesVector"
 **/
void ContentsStorageBase::refineRelease(ContentsStorageBase::RefinePtrType &refinePtr)
{
  refinesVectorMutex.lock();
  for(unsigned int i = 0; i < refinesVector.size(); i++)
  {
    if(refinesVector[i].first == refinePtr)
    {
      refinesVector[i].second = false;
      break;
    }
  }
  refinesVectorMutex.unlock();
  componentMutex.lock();
  conditionVar.signal();
  componentMutex.unlock();
  refinePtr = nullptr;
}

/** ContentsStorageBase::ContentsStorageBase class constructor.
 **/
ContentsStorageBase::ContentsStorageBase(unsigned int refinesCount)
{
  for(unsigned int i = 0; i < refinesCount; i++)
  {
    refinesVector.push_back(std::pair<RefinePtrType, bool>(RefinePtrType(new HCE::component::Refine(HCE::CT_REFINE)), false));
  }
}

/** ContentsStorageBase::ContentsStorageBase class destructor.
 **/
ContentsStorageBase::~ContentsStorageBase()
{
  bool isExit = false;
  do
  {
    isExit = true;
    refinesVectorMutex.lock();
    for(unsigned int i = 0; i < refinesVector.size(); i++)
    {
      if(refinesVector[i].second)
      {
        isExit = false;
        break;
      }
    }
    refinesVectorMutex.unlock();
    usleep(DESTRUCTOR_TIME_WAIT);
  }
  while(!isExit);
}